from sys import stderr
from numpy import (atleast_1d,arctan,sqrt,tan,sign,sin,cos,arctan2,arcsin,
                   ones,empty,zeros,radians)
from math import tau,pi

def vdist(lat1,lon1,lat2,lon2):
    """
    VDIST - Using the WGS-84 Earth ellipsoid, compute the distance between
         two points within a few millimeters of accuracy, compute forward
         azimuth, and compute backward azimuth, all using a vectorized
         version of Vincenty's algorithm.

 s = vdist(lat1,lon1,lat2,lon2)
 s,a12 = vdist(lat1,lon1,lat2,lon2)
 s,a12,a21 = vdist(lat1,lon1,lat2,lon2)

 s = distance in meters (inputs may be scalars, vectors, or matrices)
 a12 = azimuth in degrees from first point to second point (forward)
 a21 = azimuth in degrees from second point to first point (backward)
       (Azimuths are in degrees clockwise from north.)
 lat1 = GEODETIC latitude of first point (degrees)
 lon1 = longitude of first point (degrees)
 lat2, lon2 = second point (degrees)

  Original algorithm source:
  T. Vincenty, "Direct and Inverse Solutions of Geodesics on the Ellipsoid
  with Application of Nested Equations", Survey Review, vol. 23, no. 176,
  April 1975, pp 88-93.
  Available at: http://www.ngs.noaa.gov/PUBS_LIB/inverse.pdf

 Notes: (1) lat1,lon1,lat2,lon2 can be any (identical) size/shape. Outputs
           will have the same size and shape.
        (2) Error correcting code, convergence failure traps, antipodal
            corrections, polar error corrections, WGS84 ellipsoid
            parameters, testing, and comments: Michael Kleder, 2004.
        (3) Azimuth implementation (including quadrant abiguity
            resolution) and code vectorization, Michael Kleder, Sep 2005.
        (4) Vectorization is convergence sensitive; that is, quantities
            which have already converged to within tolerance are not
            recomputed during subsequent iterations (while other
            quantities are still converging).
        (5) Vincenty describes his distance algorithm as precise to within
            0.01 millimeters, subject to the ellipsoidal model.
        (6) For distance calculations, essentially antipodal points are
            treated as exactly antipodal, potentially reducing accuracy
            slightly.
        (7) Distance failures for points exactly at the poles are
            eliminated by moving the points by 0.6 millimeters.
        (8) The Vincenty distance algorithm was transcribed verbatim by
            Peter Cederholm, August 12, 2003. It was modified and
            translated to English by Michael Kleder.
            Mr. Cederholm's website is http://www.plan.aau.dk/~pce/
        (9) Distances agree with the Mapping Toolbox, version 2.2 (R14SP3)
            with a max relative difference of about 5e-9, except when the
            two points are nearly antipodal, and except when one point is
            near the equator and the two longitudes are nearly 180 degrees
            apart. This function (vdist) is more accurate in such cases.
            For example, note this difference (as of this writing):
            >>vdist(0.2,305,15,125)
            18322827.0131551
            >>distance(0.2,305,15,125,[6378137 0.08181919])
            0
       (10) Azimuths FROM the north pole (either forward starting at the
            north pole or backward when ending at the north pole) are set
            to 180 degrees by convention. Azimuths FROM the south pole are
            set to 0 degrees by convention.
       (11) Azimuths agree with the Mapping Toolbox, version 2.2 (R14SP3)
            to within about a hundred-thousandth of a degree, except when
            traversing to or from a pole, where the convention for this
            function is described in (10), and except in the cases noted
            above in (9).
       (12) No warranties; use at your own risk.
    """
#%% reshape inputs
    lat1 = atleast_1d(lat1)
    lat2 = atleast_1d(lat2)
    lon1 = atleast_1d(lon1)
    lon2 = atleast_1d(lon2)
    keepsize = lat1.shape
#%% Input check:
    if ((abs(lat1)>90) | (abs(lat2)>90)).any():
        raise ValueError('Input latitudes must be between -90 and 90 degrees, inclusive.')
#%%% Supply WGS84 earth ellipsoid axis lengths in meters:
    a = 6378137 # definitionally
    b = 6356752.31424518 # computed from WGS84 earth flattening coefficient
#%% preserve true input latitudes:
    lat1tr = lat1.copy()
    lat2tr = lat2.copy()
#%% convert inputs in degrees to radians:
    lat1 = radians(lat1)
    lon1 = radians(lon1)
    lat2 = radians(lat2)
    lon2 = radians(lon2)
#%% correct for errors at exact poles by adjusting 0.6 millimeters:
    kidx = abs(pi/2-abs(lat1)) < 1e-10;
    if kidx.any():
        lat1[kidx] = sign(lat1[kidx])*(pi/2-(1e-10))

    kidx = abs(pi/2-abs(lat2)) < 1e-10
    if kidx.any():
        lat2[kidx] = sign(lat2[kidx])*(pi/2-(1e-10))

    f = (a-b)/a
    U1 = arctan((1-f)*tan(lat1))
    U2 = arctan((1-f)*tan(lat2))
    lon1 = lon1 % tau
    lon2 = lon2 % tau
    L = abs(lon2-lon1)
    kidx = L > pi
    if kidx.any():
        L[kidx] = tau - L[kidx]

    lamb = L.copy()  # NOTE: program will fail without copy!
    lambdaold = zeros(lat1.shape)
    itercount = 0
    notdone = ones(lat1.shape,dtype=bool)
    alpha = zeros(lat1.shape)
    sigma = zeros(lat1.shape)
    cos2sigmam = zeros(lat1.shape)
    C = zeros(lat1.shape)
    warninggiven = False
    sinsigma = empty(notdone.shape)
    cossigma = empty(notdone.shape)
    while notdone.any():  # force at least one execution
        #print('iter:',itercount)
        #print(f'lambda[21752] = {lamb[21752],20}')
        itercount += 1
        if itercount > 50:
            if not warninggiven:
                print('Essentially antipodal points encountered. Precision may be reduced slightly.',file=stderr)

            lamb[notdone] = pi
            break

        lambdaold[notdone] = lamb[notdone]

        sinsigma[notdone] = sqrt(
                (cos(U2[notdone])*sin(lamb[notdone]))**2+(cos(U1[notdone])*sin(U2[notdone])-sin(U1[notdone])*
            cos(U2[notdone])*cos(lamb[notdone]))**2)

        cossigma[notdone] = (sin(U1[notdone])*sin(U2[notdone])+
            cos(U1[notdone])*cos(U2[notdone])*cos(lamb[notdone]))
        # eliminate rare imaginary portions at limit of numerical precision:
        sinsigma[notdone]= sinsigma[notdone].real
        cossigma[notdone]= cossigma[notdone].real
        
        sigma[notdone] = arctan2(sinsigma[notdone],cossigma[notdone])
        
        alpha[notdone] = (arcsin(cos(U1[notdone])*cos(U2[notdone])*
            sin(lamb[notdone])/sin(sigma[notdone])))
        
        cos2sigmam[notdone] = (cos(sigma[notdone])-2*sin(U1[notdone])*
            sin(U2[notdone])/cos(alpha[notdone])**2)
        
        C[notdone] = f/16*cos(alpha[notdone])**2*(4+f*(4-3*cos(alpha[notdone])**2))
        
        lamb[notdone] = (L[notdone]+(1-C[notdone])*f*sin(alpha[notdone])
            *(sigma[notdone]+C[notdone]*sin(sigma[notdone])*
            (cos2sigmam[notdone]+C[notdone]*cos(sigma[notdone])*
            (-1+2.*cos2sigmam[notdone]**2))))
        #print(f'then, lambda(21752) = {lamb[21752],20})
        # correct for convergence failure in the case of essentially antipodal points
        if (lamb[notdone] > pi).any():
            print('Essentially antipodal points encountered. Precision may be reduced slightly.',file=stderr)
            warninggiven = True
            lambdaold[lamb>pi] = pi
            lamb[lamb>pi] = pi

        notdone = abs(lamb-lambdaold) > 1e-12

    u2 = cos(alpha)**2*(a**2-b**2)/b**2
    A = 1+u2/16384*(4096+u2*(-768+u2*(320-175*u2)))
    B = u2/1024*(256+u2*(-128+u2*(74-47*u2)))
    deltasigma = B*sin(sigma)*(cos2sigmam+B/4*(cos(sigma)*(-1+2*
        cos2sigmam**2)-B/6*cos2sigmam*(-3+4*sin(sigma)**2)*(-3+4*cos2sigmam**2)))

    dist_m = (b*A*(sigma-deltasigma)).reshape(keepsize)

#%% From point #1 to point #2
    # correct sign of lambda for azimuth calcs:
    lamb = abs(lamb)
    kidx=sign(sin(lon2-lon1)) * sign(sin(lamb)) < 0
    lamb[kidx] = -lamb[kidx]
    numer = cos(U2)*sin(lamb)
    denom = cos(U1)*sin(U2)-sin(U1)*cos(U2)*cos(lamb)
    a12 = arctan2(numer,denom)
    kidx = a12<0
    a12[kidx]=a12[kidx] + tau
    #%% from poles
    a12[lat1tr <= -90] = 0
    a12[lat1tr >=  90] = pi
    az = (a12 * 57.2957795130823).reshape(keepsize) # to degrees

#%% From point #2 to point #1
    # correct sign of lambda for azimuth calcs:
    lamb = abs(lamb);
    kidx=sign(sin(lon1-lon2)) * sign(sin(lamb)) < 0
    lamb[kidx]=-lamb[kidx]
    numer = cos(U1)*sin(lamb)
    denom = sin(U1)*cos(U2)-cos(U1)*sin(U2)*cos(lamb)
    a21 = arctan2(numer,denom)
    kidx=a21<0;
    a21[kidx] = a21[kidx] + tau
    #%% backwards from poles:
    a21[lat2tr >= 90] = pi
    a21[lat2tr <= -90] = 0
    backaz = (a21 * 57.2957795130823).reshape(keepsize) # to degrees

    return dist_m.squeeze()[()], az.squeeze()[()], backaz.squeeze()[()]


#!/usr/bin/env python
"""
converts right ascension, declination to azimuth, elevation and vice versa



inputs:
ra_deg: numpy ndarray of right ascension values [degrees]
dec_deg: numpy ndarray of declination values [degrees]
lat_deg: scalar observer WGS84 latitude [degrees]
lon_deg: scalar observer WGS84 longitude [degrees]
dtime: UTC time of observation YYYY-mm-ddTHH:MM:SS

outputs:
az_deg: numpy ndarray of azimuth to point [degrees]
el_deg: numpy ndarray of elevation to point [degrees]

Michael Hirsch implementation of algorithms from D. Vallado
"""
from __future__ import division
from numpy import sin, cos, degrees, radians, arcsin, arctan2, atleast_1d

#
from .datetime2hourangle import datetime2sidereal


def azel2radec(az_deg, el_deg, lat_deg, lon_deg, t):
    """
    from D.Vallado Fundamentals of Astrodynamics and Applications
    p.258-259
    """
    az_deg = atleast_1d(az_deg)
    el_deg = atleast_1d(el_deg)
    lat_deg = atleast_1d(lat_deg)
    lon_deg = atleast_1d(lon_deg)

    assert az_deg.shape == el_deg.shape, 'az and el must be same shape ndarray'
    assert lat_deg.size == 1 and lon_deg.size == 1, 'need one observer and one or more  (az,el).'

    az = radians(az_deg)
    el = radians(el_deg)
    lat = radians(lat_deg)
    lon = radians(lon_deg)
    # Vallado "algorithm 28" p 268
    dec = arcsin(sin(el) * sin(lat) + cos(el) * cos(lat) * cos(az))

    lha = arctan2(-(sin(az) * cos(el)) / cos(dec),
                  (sin(el) - sin(lat) * sin(dec)) / (cos(dec) * cos(lat)))

    lst = datetime2sidereal(t, lon)  # lon, ra in RADIANS

    """ by definition right ascension \in [0,360) degrees """
    return degrees(lst - lha) % 360, degrees(dec)


def radec2azel(dtime,ra_deg,dec_deg,lat_deg,lon_deg):
    """
    from D. Vallado "Fundamentals of Astrodynamics and Applications "
       4th Edition Ch. 4.4 pg. 266-268
    """
    ra =  radians(ra_deg)
    dec = radians(dec_deg)
    lat = radians(lat_deg)
    lon = radians(lon_deg)

    lst = datetime2sidereal(dtime,lon) #RADIANS
    lha = lst - ra #Eq. 4-11 p. 267 LOCAL HOUR ANGLE

    el = arcsin(sin(lat) * sin(dec) + cos(lat) * cos(dec) * cos(lha) ) #Eq. 4-12 p. 267
    #combine Eq. 4-13 and 4-14 p. 268
    az = arctan2( -sin(lha) * cos(dec) / cos(el),
                   ( sin(dec) - sin(el) * sin(lat) ) / (cos(el)*cos(lat))  )


    return degrees(az) % 360.0, degrees(el)


#!/usr/bin/env python
"""
 Ported by Michael Hirsch to Python.
Original work by Joaquim Luis (LGPL), Michael Kleder, et al.
"""
from __future__ import division
from numpy import (absolute, sin, cos, tan, arctan2, atleast_1d,
                   radians, degrees, sign, mod, empty, pi, sqrt, tile, nan)


def vreckon(lat1, lon1, rng, azim, ellipsoid=None):
    """
     VRECKON -  Computes points at a specified azimuth and range in an
                ellipsoidal earth

                   - Using the WGS-84 Earth ellipsoid, travel a given
                     distance along
                     a given azimuth starting at a given initial point,
                     and return the
                     endpoint within a few millimeters of accuracy,
                     using Vincenty's algorithm.

     USAGE:
     lat2,lon2 = vreckon(lat1, lon1, rng, azim)
      Transmits ellipsoid definition (either as [a,b] or [a,f]) as fifth
                argument ELLIPSOID

     VARIABLES:
     lat1 = inital latitude (degrees)
     lon1 = initial longitude (degrees)
     rng  = distance (meters)
                   It can be a scalar or a vector. Latter case computes
                    a series of
                   circles (or arc circles, see azim) centered on X,Y
                    (which are scalars)
     azim = intial azimuth (degrees)
                   "azim" is a scalar or vector
     ellipsoid = two-element ellipsoid vector. Either [a b] or [a f]
                   If omitted, defaults to WGS-84
     lat2, lon2 = second point (degrees)
     a21  = reverse azimuth (degrees), at final point facing back toward
            the  intial point

     Original algorithm source:
     T. Vincenty, "Direct and Inverse Solutions of Geodesics on the
Ellipsoid
     with Application of Nested Equations", Survey Review, vol. 23, no.
176,
     April 1975, pp 88-93.
     Available at: http://www.ngs.noaa.gov/PUBS_LIB/inverse.pdf

     Notes:
     (1) The Vincenty reckoning algorithm was transcribed verbatim into
         JavaScript by Chris Veness. It was modified and translated to
Matlab
         by Michael Kleder. Mr. Veness's website is:

http://www.movable-type.co.uk/scripts/latlong-vincenty-direct.html
     (2) Error correcting code, polar error corrections, WGS84 ellipsoid
         parameters, testing, and comments by Michael Kleder.
     (3) By convention, when starting at a pole, the longitude of the
initial
         point (otherwise meaningless) determines the longitude line
along
         which to traverse, and hence the longitude of the final point.
     (4) The convention noted in (3) above creates a discrepancy with
VDIST
         when the the intial or final point is at a pole. In the VDIST
         function, when traversing from a pole, the azimuth is  0 when
         heading away from the south pole and 180 when heading away from
the
         north pole. In contrast, this VRECKON function uses the azimuth
as
        noted in (3) above when traversing away form a pole.
     (5) In testing, where the traversal subtends no more than 178
degrees,
         this function correctly inverts the VDIST function to within
0.2
         millimeters of distance, 5e-10 degrees of forward azimuth,
         and 5e-10 degrees of reverse azimuth. Precision reduces as test
         points approach antipodal because the precision of VDIST is
reduced
         for nearly antipodal points. (A warning is given by VDIST.)
     (6) Tested but no warranty. Use at your own risk.
     (7) Ver 1.0, Michael Kleder, November 2007
     (8) Ver 2.0, Joaquim Luis, September 2008

     Added ellipsoid and vectorized whenever possible.
     Also, lon2 is always converted to the [-180 180] interval
     Joaquim Luis
     $Id: $
    """

    lat1 = atleast_1d(lat1)
    lon1 = atleast_1d(lon1)
    rng = atleast_1d(rng)
    azim = atleast_1d(azim)

    assert absolute(lat1) <= 90, (
        'VRECKON: Input lat. must be between -90 and 90 deg., inclusive.')
    if lat1.size != 1 and rng.size > 1:
        raise ValueError(
            'VRECKON: Variable ranges are only allowed for a single point.')
    if ellipsoid is not None:  # An ellipsoid vector (with a & b OR a & f)
        a = ellipsoid[0]           # b = ellipsoid(2);
        # Second ellipsoid argument contains flattening instead of minor axis
        if ellipsoid[1] < 1:
            f = ellipsoid[1]
            b = a * (1 - f)
        else:               # Second ellipsoid argument contains minor axis
            f = (a - ellipsoid[1]) / a
    else:   # Supply WGS84 earth ellipsoid axis lengths in meters:
        a = 6378137                 # semimajor axis
        b = 6356752.31424518   # WGS84 earth flattening coefficient definition
        f = (a - b) / a

    lat1 = radians(lat1)            # intial latitude in radians
    lon1 = radians(lon1)            # intial longitude in radians

    # correct for errors at exact poles by adjusting 0.6 millimeters:
    kidx = absolute(pi / 2 - absolute(lat1)) < 1e-10
    lat1[kidx] = sign(lat1[kidx]) * (pi / 2 - (1e-10))

    #  Allow for multiple circles starting from the same point
    if lat1.size == 1 and lon1.size == 1 and rng.size > 1:
        lat1 = tile(lat1, rng.shape)
        lon1 = tile(lon1, rng.shape)

    if rng.size == 1:
        rng = tile(rng, azim.shape)

    if azim.size == 1:
        azim = tile(azim, rng.shape)

    assert rng.size == azim.shape[0], (
        'Range must be a scalar or vector with the same shape as azim.')

    alpha1 = radians(azim)  # inital azimuth in radians
    sinAlpha1 = sin(alpha1)
    cosAlpha1 = cos(alpha1)

    tanU1 = (1 - f) * tan(lat1)
    cosU1 = 1 / sqrt(1 + tanU1 ** 2)
    sinU1 = tanU1 * cosU1
    sigma1 = arctan2(tanU1, cosAlpha1)
    sinAlpha = cosU1 * sinAlpha1
    cosSqAlpha = 1 - sinAlpha * sinAlpha
    uSq = cosSqAlpha * (a**2 - b**2) / b**2
    A = 1 + uSq / 16384 * (4096 + uSq * (-768 + uSq * (320 - 175 * uSq)))
    B = uSq / 1024 * (256 + uSq * (-128 + uSq * (74 - 47 * uSq)))
    sigma = rng / (b * A)
    sigmaP = 2 * pi

    if sigma.size == 1:
        sinSigma = nan
        cosSigma = nan
        cos2SigmaM = nan
        while absolute(sigma - sigmaP) > 1e-12:
            cos2SigmaM = cos(2 * sigma1 + sigma)
            sinSigma = sin(sigma)
            cosSigma = cos(sigma)
            deltaSigma = (B * sinSigma *
                          (cos2SigmaM + B / 4 *
                           (cosSigma * (-1 + 2 * cos2SigmaM * cos2SigmaM) -
                            B / 6 * cos2SigmaM *
                            (-3 + 4 * sinSigma * sinSigma) *
                            (-3 + 4 * cos2SigmaM * cos2SigmaM))))
            sigmaP = sigma
            sigma = rng / (b * A) + deltaSigma
    else:
        # This part is not vectorized
        cos2SigmaM = empty(sigma.shape)
        sinSigma = empty(sigma.shape)
        cosSigma = empty(sigma.shape)

        for k in range(sigma.size):
            while (absolute(sigma[k] - sigmaP) > 1e-12).any():
                cos2SigmaM[k] = cos(2 * sigma1[k] + sigma[k])
                sinSigma[k] = sin(sigma[k])
                cosSigma[k] = cos(sigma[k])
                tmp = 2 * cos2SigmaM[k] * cos2SigmaM[k]
                deltaSigma = (B[k] * sinSigma[k] *
                              (cos2SigmaM[k] + B[k] / 4 *
                               (cosSigma[k] * (-1 + tmp) -
                                B[k] / 6 * cos2SigmaM[k] *
                                (-3 + 4 * sinSigma[k] * sinSigma[k]) *
                                  (-3 + 2 * tmp))))
                sigmaP = sigma[k]
                sigma[k] = rng[k] / (b * A[k]) + deltaSigma

    tmp = sinU1 * sinSigma - cosU1 * cosSigma * cosAlpha1
    lat2 = arctan2(sinU1 * cosSigma + cosU1 * sinSigma * cosAlpha1,
                   (1 - f) * sqrt(sinAlpha * sinAlpha + tmp**2))

    lamb = arctan2(sinSigma * sinAlpha1,
                   cosU1 * cosSigma - sinU1 * sinSigma * cosAlpha1)

    C = f / 16 * cosSqAlpha * (4 + f * (4 - 3 * cosSqAlpha))
    L = lamb - (f * (1 - C) *
                sinAlpha * (sigma + C * sinSigma *
                            (cos2SigmaM + C * cosSigma *
                             (-1 + 2 * cos2SigmaM * cos2SigmaM))))

    lon2 = degrees(lon1 + L)

    # Truncates angles into the [-pi pi] range
    # if (lon2 > pi).any():
    #    lon2 = pi*((absolute(lon2)/pi) -
    #       2*ceil(((absolute(lon2)/pi)-1)/2)) * sign(lon2)

    # lon2 = mod(lon2,360); % follow [0,360] convention
    lon2 = (lon2 + 180) % 360 - 180  # no parenthesis on RHS

    a21 = arctan2(sinAlpha, -tmp)
    a21 = 180. + degrees(a21)  # note direction reversal

    return degrees(lat2), lon2, mod(a21, 360.)



#!/usr/bin/env python
from setuptools import setup

req = ['python-dateutil','pytz','nose','numpy','astropy']

setup(name='pymap3d',
      packages=['pymap3d'],
      version = '1.1.1',
      description='Python coordinate conversions, following convention of several popular Matlab routines.',
      author = 'Michael Hirsch, Ph.D.',
      url = 'https://github.com/scienceopen/pymap3d',
      classifiers=[
      'Intended Audience :: Science/Research',
      'Development Status :: 5 - Production/Stable',
      'License :: OSI Approved :: BSD License',
      'Topic :: Scientific/Engineering :: GIS',
      'Programming Language :: Python :: 2.7',
      'Programming Language :: Python :: 3.4',
      'Programming Language :: Python :: 3.5',
      'Programming Language :: Python :: 3.6',
      ],
      install_requires=req,
	  )




#!/usr/bin/env python
from __future__ import division
from dateutil.parser import parse
from datetime import datetime
from numpy import (sin, cos, tan, sqrt, radians, arctan2, hypot, degrees, mod,
                   atleast_2d, atleast_1d, empty_like, array, column_stack)
try:
    from astropy.time import Time
    from astropy import units as u
    from astropy.coordinates import Angle,SkyCoord, EarthLocation, AltAz, ICRS
except ImportError:
    pass
"""
Michael Hirsch ported and adaptation from
 GNU Octave Mapping Toolbox by
  Copyright (c) 2013, Sandeep V. Mahanthi
Copyright (c) 2013, Felipe G. Nievinski

Input/output: units are METERS and DEGREES.
boolean deg=True means degrees

For most functions you can input Numpy arrays of any shape,
except as noted in the functions

see test.py for example uses.
"""
class EarthEllipsoid:

    def __init__(self):
        self.a = 6378137.  # semi-major axis [m]
        self.f = 1 / 298.2572235630  # flattening
        self.b = self.a * (1 - self.f)  # semi-minor axis
#%% to AER (azimuth, elevation, range)
def ecef2aer(x, y, z, lat0, lon0, h0, ell=EarthEllipsoid(), deg=True):
    xEast, yNorth, zUp = ecef2enu(x, y, z, lat0, lon0, h0, ell, deg=deg)
    return enu2aer(xEast, yNorth, zUp, deg=deg)


def eci2aer(eci, lat0, lon0, h0, t):
    ecef = eci2ecef(eci, t)
    return ecef2aer(ecef[:, 0], ecef[:, 1], ecef[:, 2], lat0, lon0, h0)


def enu2aer(e, n, u, deg=True):
    """
    input: east, north, up [m]
    """
    r = hypot(e, n)
    slantRange = hypot(r, u)
    elev = arctan2(u, r)
    az = mod(arctan2(e, n), 2 * arctan2(0, -1))
    if deg:
        return degrees(az), degrees(elev), slantRange
    else:
        return az, elev, slantRange  # radians


def geodetic2aer(lat, lon, h, lat0, lon0, h0, ell=EarthEllipsoid(), deg=True):
    e, n, u = geodetic2enu(lat, lon, h, lat0, lon0, h0, ell, deg=deg)
    return enu2aer(e, n, u, deg=deg)


def ned2aer(n, e, d, deg=True):
    return enu2aer(e, n, -d, deg=deg)
#%% to ECEF
def aer2ecef(az, el, srange, lat0, lon0, alt0, ell=EarthEllipsoid(), deg=True):
    """
     Input: az,el; lat0,lon0 [degrees]   srange, alt0 [meters]

     output: ECEF x,y,z  [meters]

    if you specify NaN for srange, return value z will be NaN
    """
    # Origin of the local system in geocentric coordinates.
    x0, y0, z0 = geodetic2ecef(lat0, lon0, alt0, ell, deg=deg)
    # Convert Local Spherical AER to ENU
    e1, n1, u1 = aer2enu(az, el, srange, deg=deg)
    # Rotating ENU to ECEF
    dx, dy, dz = _enu2uvw(e1, n1, u1, lat0, lon0, deg=deg)
    # Origin + offset from origin equals position in ECEF
    return x0 + dx, y0 + dy, z0 + dz


def eci2ecef(eci, t):
    """
    input t is either a datetime or float in radians

    """
    t = atleast_1d(t)
    if isinstance(t[0],str): #don't just ram in in case it's float
        t = str2dt(t)

    if isinstance(t[0], datetime):
        gst = Time(t).sidereal_time('apparent', 'greenwich').radian
    elif isinstance(t[0],float):
        gst = t
    else:
        raise TypeError('eci2ecef: time must be datetime or radian float')

    assert isinstance(gst[0], float)  # must be in radians!

    eci = atleast_2d(eci)
    N, trip = eci.shape
    if eci.ndim > 2 or trip != 3:
        raise TypeError('eci triplets must be shape (N,3)')
    """ported from:
    https://github.com/dinkelk/astrodynamics/blob/master/rot3.m
    """
    ecef = empty_like(eci)

    for i in range(N):
        ecef[i, :] = _rottrip(gst[i]).dot(eci[i, :])
    return ecef


def enu2ecef(e1, n1, u1, lat0, lon0, alt0, ell=EarthEllipsoid(), deg=True):
    x0, y0, z0 = geodetic2ecef(lat0, lon0, alt0, ell, deg=deg)
    dx, dy, dz = _enu2uvw(e1, n1, u1, lat0, lon0, deg=deg)
    return x0 + dx, y0 + dy, z0 + dz


def geodetic2ecef(lat, lon, alt, ell=EarthEllipsoid(), deg=True):
    if deg:
        lat = radians(lat)
        lon = radians(lon)
    # radius of curvature of the prime vertical section
    N = get_radius_normal(lat, ell)
    # Compute cartesian (geocentric) coordinates given  (curvilinear) geodetic
    # coordinates.
    x = (N + alt) * cos(lat) * cos(lon)
    y = (N + alt) * cos(lat) * sin(lon)
    z = (N * (ell.b / ell.a)**2 + alt) * sin(lat)
    return x, y, z


def ned2ecef(n, e, d, lat0, lon0, h0, ell=EarthEllipsoid(), deg=True):
    return enu2ecef(e, n, -d, lat0, lon0, h0, ell, deg=deg)
#%% to ECI
def aer2eci(az, el, srange, lat0, lon0, h0, t, ell=EarthEllipsoid(), deg=True):
    x, y, z = aer2ecef(az, el, srange, lat0, lon0, h0, ell, deg)
    return ecef2eci(column_stack((x, y, z)), t)


def ecef2eci(ecef, t):
    """
    input t is either a datetime or float in radians

    """
    t = atleast_1d(t)
    if isinstance(t[0],str): #don't just ram in in case it's float
        t = str2dt(t)

    if isinstance(t[0], datetime):
        gst = Time(t).sidereal_time('apparent', 'greenwich').radian
    elif isinstance(t[0],float):
        gst = t
    else:
        raise TypeError('eci2ecef: time must be datetime or radian float')

    assert isinstance(gst[0], float)  # must be in radians!

    ecef = atleast_2d(ecef)
    N, trip = ecef.shape
    if ecef.ndim > 2 or trip != 3:
        raise TypeError('ecef triplets must be shape (N,3)')
    """ported from:
    https://github.com/dinkelk/astrodynamics/blob/master/rot3.m
    """
    eci = empty_like(ecef)
    for i in range(N):
        eci[i, :] = _rottrip(gst[i]).T.dot(
            ecef[i, :])  # this one is transposed
    return eci
#%% to ENU
def aer2enu(az, el, srange, deg=True):
    """
    input: azimuth, elevation [deg]    slant range [m]
    output: East, North, Up [m]
    """
    if deg:
        el = radians(el)
        az = radians(az)

    r = srange * cos(el)
    return r * sin(az), r * cos(az), srange * sin(el)


def ecef2enu(x, y, z, lat0, lon0, h0, ell=EarthEllipsoid(), deg=True):
    x0, y0, z0 = geodetic2ecef(lat0, lon0, h0, ell, deg=deg)
    return _uvw2enu(x - x0, y - y0, z - z0, lat0, lon0, deg=deg)


def geodetic2enu(lat, lon, h, lat0, lon0, h0, ell=EarthEllipsoid(), deg=True):
    x1, y1, z1 = geodetic2ecef(lat, lon, h, ell, deg=deg)
    x2, y2, z2 = geodetic2ecef(lat0, lon0, h0, ell, deg=deg)
    dx = x1 - x2
    dy = y1 - y2
    dz = z1 - z2
    return _uvw2enu(dx, dy, dz, lat0, lon0, deg=deg)
#%% to geodetic
def aer2geodetic(az, el, srange, lat0, lon0, alt0, deg=True):
    """
    Input: az,el; lat0,lon0 [degrees]   srange, alt0 [meters]

    output: WGS84 lat,lon [degrees]  altitude above spheroid  [meters]
    """
    x, y, z = aer2ecef(az, el, srange, lat0, lon0, alt0, deg=deg)
    return ecef2geodetic(x, y, z, deg=deg)


def ecef2geodetic(x, y=None, z=None, ell=EarthEllipsoid(), deg=True):
    if y is None:
        x, y, z = _depack(x)
    """Algorithm is based on
    http://www.astro.uni.torun.pl/~kb/Papers/geod/Geod-BG.htm
    This algorithm provides a converging solution to the latitude
equation
    in terms of the parametric or reduced latitude form (v)
    This algorithm provides a uniform solution over all latitudes as it
does
    not involve division by cos(phi) or sin(phi)
    """
    ea = ell.a
    eb = ell.b
    rad = hypot(x, y)
# Constant required for Latitude equation
    rho = arctan2(eb * z, ea * rad)
# Constant required for latitude equation
    c = (ea**2 - eb**2) / hypot(ea * rad, eb * z)
# Starter for the Newtons Iteration Method
    vnew = arctan2(ea * z, eb * rad)
# Initializing the parametric latitude
    v = 0
    count = 0
    while (v != vnew).any() and count < 5:
        v = vnew.copy()
#%% Newtons Method for computing iterations
        vnew = v - ((2 * sin(v - rho) - c * sin(2 * v)) /
                    (2 * (cos(v - rho) - c * cos(2 * v))))
        count += 1

#%% Computing latitude from the root of the latitude equation
    lat = arctan2(ea * tan(vnew), eb)
    # by inspection
    lon = arctan2(y, x)

    alt = ((rad - ea * cos(vnew)) * cos(lat)) + \
        ((z - eb * sin(vnew)) * sin(lat))

    if deg:
        return degrees(lat), degrees(lon), alt
    else:
        return lat, lon, alt  # radians
"""
this is from PySatel and gives same result to EIGHT decimal places
def cbrt(x):
    if x >= 0:
        return pow(x, 1.0/3.0)
    else:
        return -pow(abs(x), 1.0/3.0)

def ecef2geodetic(x, y, z, ell=EarthEllipsoid(),deg=True):
    a = ell.a; b = ell.b
    esq = 6.69437999014*0.001
    e1sq = 6.73949674228*0.001
    r = hypot(x,y)
    Esq = a**2 - b**2
    F = 54 * b**2 * z**2
    G = r**2 + (1 - esq)* z**2 - esq*Esq
    C = (esq**2 *F* r**2)/(pow(G, 3))
    S = cbrt(1 + C + sqrt(C**2 + 2*C))
    P = F/(3* pow((S + 1/S + 1), 2)*G**2)
    Q = sqrt(1 + 2* esq**2 *P)
    r_0 =  -(P*esq*r)/(1 + Q) + sqrt(0.5* a**2 *(1 + 1.0/Q) - \
           P*(1 - esq)*z**2/(Q*(1 + Q)) - 0.5*P* r**2)
    U = sqrt(pow((r - esq*r_0), 2) + z**2)
    V = sqrt(pow((r - esq*r_0), 2) + (1 - esq)* z**2)
    Z_0 = b**2 *z/(a*V)
    alt = U*(1 - b**2/(a*V))
    lat = arctan((z + e1sq*Z_0)/r)
    lon = arctan2(y, x)

    if deg:
        return degrees(lat),degrees(lon),alt
    else:
        return lat, lon, alt #radians
"""

def eci2geodetic(eci, t):
    """
    inputs:
    -------
    eci/ecef: a Nx3 vector of x,y,z triplets in the eci or ecef system [meters]
    t : length N vector of datetime OR greenwich sidereal time angle [radians].

    Note: Conversion is idealized: doesn't consider nutations, perterbations,
    etc. like the IAU-76/FK5 or IAU-2000/2006 model-based conversions
    from ECI to ECEF
    """

    """ a.k.a. eci2lla() """
    ecef = eci2ecef(eci, t)
    return ecef2geodetic(ecef[:, 0], ecef[:, 1], ecef[:, 2])


def enu2geodetic(e, n, u, lat0, lon0, h0,  ell=EarthEllipsoid(), deg=True):
    x, y, z = enu2ecef(e, n, u, lat0, lon0, h0, ell, deg=deg)
    return ecef2geodetic(x, y, z, ell, deg=deg)


def ned2geodetic(n, e, d, lat0, lon0, h0, ell=EarthEllipsoid(), deg=True):
    x, y, z = enu2ecef(e, n, -d, lat0, lon0, h0,  ell, deg=deg)
    return ecef2geodetic(x, y, z, ell, deg=deg)
#%% to NED

def aer2ned(az, elev, slantRange, deg=True):
    e, n, u = aer2enu(az, elev, slantRange, deg=deg)
    return n, e, -u


def ecef2ned(x, y, z, lat0, lon0, h0, ell=EarthEllipsoid(), deg=True):
    e, n, u = ecef2enu(x, y, z, lat0, lon0, h0, ell, deg=deg)
    return n, e, -u

def geodetic2ned(lat, lon, h, lat0, lon0, h0, ell=EarthEllipsoid(), deg=True):
    e, n, u = geodetic2enu(lat, lon, h, lat0, lon0, h0, ell, deg=deg)
    return n, e, -u

#%% shared functions

def get_radius_normal(lat_radians, ell):
    a = ell.a
    b = ell.b
    return a**2 / sqrt(
        a**2 * (cos(lat_radians))**2 + b**2 *
        (sin(lat_radians))**2)



def str2dt(t):
    """
    output: datetime
    """

    t = atleast_1d(t)
    if isinstance(t[0],str):
        t = [parse(T) for T in t]

    assert isinstance(t[0],datetime), 'did not convert {} to datetime'.format(type(t[0]))

    return t
#%% internal use
def _depack(x0):
    m, n = x0.shape

    assert x0.ndim == 2 and (m==3 or n==3),'I expect Nx3 or 3XN triplets'

    if m == 3:  # 3xN triplets
        x = x0[0, :]
        y = x0[1, :]
        z = x0[2, :]
    elif n == 3:  # Nx3 triplets
        x = x0[:, 0]
        y = x0[:, 1]
        z = x0[:, 2]
    else:
        raise TypeError('I expect an Nx3 or 3xN input of x,y,z')
    return x, y, z


def _rottrip(ang):
    ang = ang.squeeze()
    if ang.size > 1:
        raise TypeError('only one angle allowed at a time')
    """ported from:
    https://github.com/dinkelk/astrodynamics/blob/master/rot3.m
    """
    return array([[cos(ang),  sin(ang), 0],
                  [-sin(ang), cos(ang), 0],
                  [0,         0,        1]])


def _enu2uvw(east, north, up, lat0, lon0, deg=True):
    if deg:
        lat0 = radians(lat0)
        lon0 = radians(lon0)
    t = cos(lat0) * up - sin(lat0) * north
    w = sin(lat0) * up + cos(lat0) * north

    u = cos(lon0) * t - sin(lon0) * east
    v = sin(lon0) * t + cos(lon0) * east
    return u, v, w


def _uvw2enu(u, v, w, lat0, lon0, deg):
    if deg:
        lat0 = radians(lat0)
        lon0 = radians(lon0)
    t = cos(lon0) * u + sin(lon0) * v
    East = -sin(lon0) * u + cos(lon0) * v
    Up = cos(lat0) * t + sin(lat0) * w
    North = -sin(lat0) * t + cos(lat0) * w
    return East, North, Up

#%% azel radec
def azel2radec(az_deg, el_deg, lat_deg, lon_deg, t):

    t = str2dt(t)

    obs = EarthLocation(lat=lat_deg * u.deg, lon=lon_deg * u.deg)
    direc = AltAz(location=obs, obstime=Time(t),
                  az=az_deg * u.deg, alt=el_deg * u.deg)
    sky = SkyCoord(direc.transform_to(ICRS()))

    return sky.ra.deg, sky.dec.deg

def radec2azel(ra_deg,dec_deg,lat_deg,lon_deg, t):
#%% input trapping
    t = str2dt(t)
    lat_deg = atleast_1d(lat_deg); lon_deg = atleast_1d(lon_deg)
    ra_deg = atleast_1d(ra_deg); dec_deg = atleast_1d(dec_deg)

    assert lat_deg.size == 1 & lon_deg.size ==1, 'radec2azel is designed for one observer and one or more points (ra,dec).'
    assert ra_deg.shape == dec_deg.shape, 'ra and dec must be the same shape ndarray'

    obs = EarthLocation(lat=lat_deg*u.deg, lon=lon_deg*u.deg)
    points = SkyCoord(Angle(ra_deg, unit=u.deg), Angle(dec_deg, unit=u.deg), equinox='J2000.0')
    altaz = points.transform_to(AltAz(location=obs, obstime=Time(t)))

    return altaz.az.degree, altaz.alt.degree
